/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 28/03/2007 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECore.h>
#include <SCE/interface/SCEScene.h>

int SCE_Init_Scene (void)
{
    return SCE_OK;
}
void SCE_Quit_Scene (void)
{
}


static void SCE_Scene_DefaultRenderFunc (SCE_SScene *scene, void *unused);

/* ajoute le 03/04/2008 */
void SCE_Scene_InitInfos (SCE_SSceneInfos *infos)
{
    infos->camera = NULL;
}
/* ajoute le 23/02/2008 */
/* revise le 06/04/2008 */
void SCE_Scene_Init (SCE_SScene *scene)
{
    int default_states[] = {1};
    SCE_Scene_InitInfos (&scene->infos);
    scene->statemanager = NULL;
    scene->render = SCE_Scene_DefaultRenderFunc;
    scene->rendertarget = NULL;
    scene->rootnode = NULL;
    scene->models = NULL;
    scene->lights = NULL;
    scene->cameras = NULL;
    scene->frustum_culling = SCE_FALSE;
    scene->use_skybox = SCE_FALSE;

    scene->states = default_states;
    scene->num_states = 1;
}

/* ajoute le 23/02/2008 */
/* revise le 20/03/2008 */
SCE_SScene* SCE_Scene_Create (void)
{
    SCE_SCamera *cam = NULL;
    SCE_SScene *scene = NULL;
    SCE_btstart ();
    if (!(scene = SCE_malloc (sizeof *scene)))
        Logger_LogSrc ();
    else
    {
        SCE_Scene_Init (scene);
#define SCE_SCENE_ASSERT(c)\
        if (c)\
        {\
            SCE_Scene_Delete (scene);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return NULL;\
        }

        SCE_SCENE_ASSERT (!(scene->statemanager = SCE_State_CreateManager ()))
        SCE_SCENE_ASSERT (!(scene->models = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->lights = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->cameras = SCE_List_Create (NULL)))

        SCE_SCENE_ASSERT (!(cam = SCE_Camera_Create ()))
        SCE_SCENE_ASSERT (SCE_Scene_AddCamera (scene, cam) < 0)
        if (!(scene->rootnode = SCE_Node_Create ()))
        {
            SCE_Scene_Delete (scene), scene = NULL;
            Logger_LogSrc ();
        }
#undef SCE_SCENE_ASSERT
    }
    SCE_btend ();
    return scene;
}
/* ajoute le 23/02/2008 */
void SCE_Scene_Delete (SCE_SScene *scene)
{
    if (scene)
    {
        SCE_Node_Delete (scene->rootnode);
        SCE_List_Delete (scene->cameras);
        SCE_List_Delete (scene->lights);
        SCE_List_Delete (scene->models);
        SCE_State_DeleteManager (scene->statemanager);
        SCE_free (scene);
    }
}


/* ajoute le 20/03/2008 */
void SCE_Scene_EnableFrustumCulling (SCE_SScene *scene)
{
    scene->frustum_culling = SCE_TRUE;
}
/* ajoute le 20/03/2008 */
void SCE_Scene_DisableFrustumCulling (SCE_SScene *scene)
{
    scene->frustum_culling = SCE_FALSE;
}

/* ajoute le 03/04/2008 */
void SCE_Scene_EnableSkybox (SCE_SScene *scene)
{
    scene->use_skybox = SCE_TRUE;
}
/* ajoute le 03/04/2008 */
void SCE_Scene_DisableSkybox (SCE_SScene *scene)
{
    scene->use_skybox = SCE_FALSE;
}

/* ajoute le 20/03/2008 */
void SCE_Scene_GenerateBoundingsBoxes (SCE_SScene *scene)
{
    SCE_SListIterator *it = NULL;
    SCE_List_ForEach (it, scene->models)
        SCE_Model_GenerateBoundingBox (SCE_List_GetData (it));
}


/* ajoute le 26/02/2008 */
SCE_SNode* SCE_Scene_GetRootNode (SCE_SScene *scene)
{
    return scene->rootnode;
}

/* ajoute le 20/03/2008 */
SCE_SCamera* SCE_Scene_GetCamera (SCE_SScene *scene, unsigned int index)
{
    SCE_SListIterator *it = NULL;
    it = SCE_List_GetIterator (scene->cameras, index);
    return SCE_List_GetData (it);
}

/* ajoute le 26/02/2008 */
int SCE_Scene_AddNode (SCE_SScene *scene, SCE_SNode *node)
{
    if (SCE_Node_Attach (scene->rootnode, node) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 26/02/2008 */
/* revise le 07/04/2008 */
int SCE_Scene_AddModel (SCE_SScene *scene, SCE_SModel *model)
{
    if (SCE_List_AddBack (scene->models, model) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    if (!SCE_Model_HaveRenderFunc (model))
        SCE_Model_SetRenderFunc (model, SCE_Scene_RenderModel, scene);
    return SCE_OK;
}

/* ajoute le 06/04/2008 */
/* revise le 07/04/2008 */
int SCE_Scene_AddSkybox (SCE_SScene *scene, SCE_SModel *skybox)
{
    if (SCE_List_AddFront (scene->models, skybox) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    if (!SCE_Model_HaveRenderFunc (skybox))
        SCE_Model_SetRenderFunc (skybox, SCE_Scene_RenderSkybox, scene);
    scene->use_skybox = SCE_TRUE; /* hopla */
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
int SCE_Scene_AddLight (SCE_SScene *scene, SCE_SLight *light)
{
    if (SCE_List_AddBack (scene->lights, light) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
int SCE_Scene_AddCamera (SCE_SScene *scene, SCE_SCamera *cam)
{
    if (SCE_List_AddBack (scene->cameras, cam) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}


/* propositions de fonctions de traitement du rendu */

/* ajoute le 31/03/2008 */
void SCE_Scene_SetRenderFunc (SCE_SScene *scene, SCE_FRenderSceneFunc func)
{
    scene->render = func;
}

/* ajoute le 31/03/2008 */
void SCE_Scene_SetRenderTarget (SCE_SScene *scene, SCE_STexture *target)
{
    scene->rendertarget = target;
}
SCE_STexture* SCE_Scene_GetRenderTarget (SCE_SScene *scene)
{
    return scene->rendertarget;
}

#if 0
/* configure le mode de rendu suivant la cible de rendu
   (defini une boucle pour une cubemap par exemple) */
void SCE_Scene_ConfigureCameraForRenderTarget (target)
{

}
#endif


/* ajoute le 02/04/2008 */
void SCE_Scene_ActiveLightsForModel (SCE_SScene *scene, SCE_SModel *mdl)
{
    SCE_SListIterator *i = NULL;
    SCE_SLight *light = NULL;
    SCE_SBoundingBox *box = NULL;
    SCE_TVector3 vec1, vec2;
    float length;
    float *p = NULL;

    box = SCE_Model_GetBoundingBox (mdl);

    /* obtention de donnees necessaire au calcul des lumieres */
    SCE_BoundingBox_GetCenterv (box, vec1);
    SCE_Vector3_Copy (vec2, &(SCE_BoundingBox_GetPoints (box)[0]));
    length = SCE_Vector3_Distance (vec1, vec2);

    /* activation des lumieres */
    SCE_List_ForEach (i, scene->lights)
    {
        light = SCE_List_GetData (i);

        /* on n'active la lumiere que si le modele est a sa portee */
        p = SCE_Light_GetPosition (light);
        if (SCE_Vector3_Distance (p, vec1) <= SCE_Light_GetRadius(light)+length)
            SCE_Light_Use (light);
    }
}

/* ajoute le 03/04/2008 */
void SCE_Scene_PrepareRendering (SCE_SScene *scene)
{
    if (scene->use_skybox)
        SCE_CClear (GL_DEPTH_BUFFER_BIT);
    else
        SCE_CClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/* ajoute le 05/04/2008 */
int SCE_Scene_ModelInFrustum (SCE_SScene *scene, SCE_SModel *mdl)
{
    int render;
    SCE_SBoundingBox *box = SCE_Model_GetBoundingBox (mdl);
    SCE_SCamera *cam = scene->infos.camera;

    /* application de la matrice a la bouding box */
    SCE_BoundingBox_Push (box, SCE_Node_GetFinalMatrix(SCE_Model_GetNode(mdl)));
    /* on verifie si le modele est dans le frustum */
    render = SCE_Frustum_BoundingBoxIn (SCE_Camera_GetFrustum (cam),
                                        SCE_Model_GetBoundingBox (mdl));
    SCE_BoundingBox_Pop (box);
    return render;
}


/* ajoute le 31/03/2008 */
/* revise le 06/04/2008 */
void SCE_Scene_RenderModel (SCE_SModel *mdl, void *s)
{
    if (((SCE_SScene*)s)->frustum_culling)
    {
        if (SCE_Scene_ModelInFrustum (s, mdl) == SCE_FRUSTUM_OUT)
            return;
    }

    /* activation des lumieres */
    SCE_Scene_ActiveLightsForModel (s, mdl);
    /* rendu du modele */
    /* on est obliges de passer par cette fonction directement... */
    SCE_Model_RenderFunc (mdl, NULL);
    SCE_Light_Use (NULL);
}

/* ajoute le 03/04/2008 */
/* ajoute le 06/04/2008 */
void SCE_Scene_RenderSkybox (SCE_SModel *mdl, void *s)
{
    SCE_SCamera *cam = ((SCE_SScene*)s)->infos.camera;
    SCE_TVector3 campos = {0., 0., 0.};
    SCE_TMatrix4 cmat;
    float *mat = SCE_Node_GetMatrix (SCE_Model_GetNode (mdl));

    SCE_Matrix4_Copy (cmat, SCE_Camera_GetView (cam));
    SCE_Matrix4_Inverse (cmat);
    SCE_Matrix4_GetTranslation (cmat, campos);
    SCE_Matrix4_Copy (cmat, mat); /* sauvegarde de la matrice */
    SCE_Matrix4_MulTranslatev (mat, campos);
    /* glPushAttribs */
    SCE_CSetState (GL_DEPTH_TEST, SCE_FALSE);
    glDepthMask (0); /* TODO: hack laid */
    SCE_Model_RenderFunc (mdl, NULL);
    glDepthMask (1);
    SCE_CSetState (GL_DEPTH_TEST, SCE_TRUE);
    SCE_Matrix4_Copy (mat, cmat); /* restauration de la matrice */
}

/* ajoute le 19/03/2008 */
/* revise le 23/03/2008 */
static void SCE_Scene_DefaultRenderFunc (SCE_SScene *scene, void *unused)
{
    SCE_SListIterator *i = NULL, *j = NULL;
    SCE_SModel *mdl = NULL;
    SCE_SCamera *cam = NULL;
    int state;

    /* pour chaque etat */
    SCE_Scene_ForEachState (state, scene)
    {
        SCE_Scene_PushState (state, scene);

        SCE_Scene_PrepareRendering (scene);

        /* pour chaque camera active */
        SCE_List_ForEach (i, scene->cameras)
        {
            cam = SCE_List_GetData (i);

            /* mise a jour des infos de rendu */
            scene->infos.camera = cam;

            /* activation de la camera et calcul du frustum (si necessaire) */
            SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
            SCE_CPushMatrix ();
            SCE_Camera_Use (cam);
            if (scene->frustum_culling)
                SCE_Camera_UpdateFrustum (cam);

            /* assignation de la matrice camera au gestionnaire de lumieres */
            SCE_Light_SetCameraMatrix (SCE_Camera_GetView (cam));

            /* pour chaque modele de la scene */
            SCE_List_ForEach (j, scene->models)
            {
                mdl = SCE_List_GetData (j);
                SCE_Model_Render (mdl);
            }

            SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
            SCE_CPopMatrix ();
            break; /* TODO: relativement temporaire */
        }
        SCE_CFlush ();
        SCE_Scene_PopState (scene);
    }
}


/* ajoute le 28/03/2008 */
void SCE_Scene_Render (SCE_SScene *scene, void *param)
{
    scene->render (scene, param);
}
